use crate::{
    errors::NoUserInfo,
    markups::deletion_markup,
    models::{DecryptedAccount, User},
    Handler, MainDialogue, State,
};
use futures::{stream, StreamExt};
use itertools::Itertools;
use parking_lot::Mutex;
use sea_orm::prelude::*;
use std::sync::Arc;
use teloxide::{adaptors::Throttle, prelude::*};
use tokio::task::spawn_blocking;

/// Ecryptes the account and adds it to the database
/// If any of these steps fail, the account name will be added to the failed vector
#[inline]
async fn encrypt_account(
    account: DecryptedAccount,
    user_id: u64,
    db: &DatabaseConnection,
    master_pass: Arc<str>,
    failed: &Mutex<&mut Vec<String>>,
) {
    let name = account.name.clone();
    match spawn_blocking(move || account.into_account(user_id, &master_pass)).await {
        Ok(Ok(account)) => match account.insert(db).await {
            Ok(_) => (),
            Err(_) => failed.lock().push(name),
        },
        _ => failed.lock().push(name),
    }
}

/// Gets the master password, encryptes and adds the accounts to the DB
async fn get_master_pass(
    bot: Throttle<Bot>,
    msg: Message,
    db: DatabaseConnection,
    dialogue: MainDialogue,
    master_pass: String,
    user: User,
) -> crate::Result<()> {
    let user_id = msg.from().ok_or(NoUserInfo)?.id.0;
    let mut failed = Vec::new();

    {
        let failed = Mutex::new(&mut failed);
        let master_pass: Arc<str> = master_pass.into();

        stream::iter(user.accounts)
            .for_each_concurrent(3, |account| {
                encrypt_account(account, user_id, &db, master_pass.clone(), &failed)
            })
            .await;
    }

    let message = if failed.is_empty() {
        "Success".to_owned()
    } else {
        format!(
            "Failed to create the following accounts:\n{}",
            failed.into_iter().format("\n")
        )
    };
    bot.send_message(msg.chat.id, message)
        .reply_markup(deletion_markup())
        .await?;
    dialogue.exit().await?;
    Ok(())
}

/// Downloads and parses the json and asks for the master password
async fn get_document(
    bot: Throttle<Bot>,
    msg: Message,
    _: DatabaseConnection,
    dialogue: MainDialogue,
    user: User,
) -> crate::Result<()> {
    let previous = bot
        .send_message(msg.chat.id, "Send master password")
        .await?;
    dialogue
        .update(State::GetMasterPass(Handler::new(
            move |bot, msg, db, dialogue, master_pass| {
                get_master_pass(bot, msg, db, dialogue, master_pass, user)
            },
            previous,
        )))
        .await?;
    Ok(())
}

/// Handles /import command
pub async fn import(bot: Throttle<Bot>, msg: Message, dialogue: MainDialogue) -> crate::Result<()> {
    let previous = bot
        .send_message(
            msg.chat.id,
            "Send a json document with the same format as created by /export",
        )
        .await?;
    dialogue
        .update(State::GetUser(Handler::new(get_document, previous)))
        .await?;
    Ok(())
}
